from functools import partial
from types import ModuleType
from typing import List, Union

import pytest
from caio import python_aio_asyncio

from aiofile import AIOFile, async_open


try:
    from caio import thread_aio_asyncio
except ImportError:
    thread_aio_asyncio = None       # type: ignore

try:
    from caio import linux_aio_asyncio
except ImportError:
    linux_aio_asyncio = None        # type: ignore


IMPLEMENTATIONS: List[Union[ModuleType, None]] = list(
    filter(
        None, [
            linux_aio_asyncio,
            thread_aio_asyncio,
            python_aio_asyncio,
        ],
    ),
)

IMPLEMENTATIONS.insert(0, None)

IMPLEMENTATION_NAMES: List[str] = list(
    map(
        lambda x: x.__name__ if x is not None else "default",
        IMPLEMENTATIONS
    )
)


@pytest.fixture(params=IMPLEMENTATIONS, ids=IMPLEMENTATION_NAMES)
async def aio_context(request, event_loop):
    if request.param is None:
        yield None
        return

    async with request.param.AsyncioContext(loop=event_loop) as context:
        yield context


@pytest.fixture
def aio_file_maker(aio_context):
    return partial(AIOFile, context=aio_context)


@pytest.fixture(name="async_open")
def async_open_maker(aio_context):
    return partial(async_open, context=aio_context)



import importlib.metadata
from email.message import Message
from email.utils import parseaddr
from typing import cast


package_metadata = cast(Message, importlib.metadata.metadata("aiofile"))

_author_email_raw = package_metadata.get("Author-email", "")
_author_name, _author_email_addr = parseaddr(_author_email_raw)

__author__ = package_metadata.get("Author", _author_name)
__version__ = package_metadata["Version"]
author_info = [(__author__, _author_email_addr or _author_email_raw)]
package_info = package_metadata.get("Summary", "")
package_license = package_metadata.get(
    "License-Expression", package_metadata.get("License", ""),
)
project_home = next(
    (
        url.split(",")[1].strip()
        for url in package_metadata.get_all("Project-URL", [])
        if "homepage" in url.lower()
    ),
    "",
)
team_email = _author_email_addr or _author_email_raw
version_info = tuple(map(int, __version__.split(".")))



from .aio import AIOFile
from .utils import (
    BinaryFileWrapper, FileIOCloner, FileIOWrapperBase, LineReader, Reader,
    TextFileWrapper, Writer, async_open, clone,
)
from .version import (
    __author__, __version__, author_info, package_info, package_license,
    project_home, team_email, version_info,
)


__all__ = (
    "AIOFile",
    "BinaryFileWrapper",
    "FileIOCloner",
    "FileIOWrapperBase",
    "LineReader",
    "Reader",
    "TextFileWrapper",
    "Writer",
    "__author__",
    "__version__",
    "async_open",
    "author_info",
    "clone",
    "package_info",
    "package_license",
    "project_home",
    "team_email",
    "version_info",
)



import asyncio
import hashlib
import json
import os
from base64 import b64encode
from io import BytesIO
from pathlib import Path
from random import shuffle
from typing import cast
from unittest.mock import Mock, call
from uuid import uuid4

import caio
import pytest

from aiofile import AIOFile, clone
from aiofile.utils import (
    BinaryFileWrapper, LineReader, Reader, TextFileWrapper, Writer,
)

from .impl import split_by


class CoroutineMock(Mock):
    async def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)


@pytest.fixture
def temp_file(tmp_path):
    path = str(tmp_path / "file.bin")
    with open(path, "wb"):
        pass

    return path


@pytest.fixture
def uuid(tmp_path):
    return str(uuid4())


async def test_read(aio_file_maker, temp_file, uuid):
    with open(temp_file, "w") as f:
        f.write(uuid)

    aio_file = await aio_file_maker(temp_file, "r")

    data = await aio_file.read()

    assert data == uuid


async def test_read_write(aio_file_maker, temp_file, uuid):
    r_file = await aio_file_maker(temp_file, "r")
    w_file = await aio_file_maker(temp_file, "w")

    await w_file.write(uuid)
    await w_file.fsync()

    data = await r_file.read()

    assert data == uuid


async def test_read_write_pathlib(aio_file_maker, temp_file, uuid):
    r_file = await aio_file_maker(Path(temp_file), "r")
    w_file = await aio_file_maker(Path(temp_file), "w")

    await w_file.write(uuid)
    await w_file.fsync()

    data = await r_file.read()

    assert data == uuid


@pytest.mark.parametrize("count", [2, 3, 5, 10, 20, 100])
async def test_read_offset(count, aio_file_maker, temp_file, uuid):
    with open(temp_file, "w") as f:
        for _ in range(count):
            f.write(uuid)

    aio_file = await aio_file_maker(temp_file, "r")

    data = await aio_file.read(
        offset=len(uuid),
        size=len(uuid),
    )

    assert data == uuid


@pytest.mark.parametrize("count", [2, 3, 5, 10, 20, 100])
async def test_read_write_offset(count, aio_file_maker, temp_file, uuid):
    r_file = await aio_file_maker(temp_file, "r")
    w_file = await aio_file_maker(temp_file, "w")

    for i in range(count):
        await w_file.write(uuid, offset=i * len(uuid))

    await w_file.fsync()

    data = await r_file.read(
        offset=len(uuid),
        size=len(uuid),
    )

    assert data == uuid


@pytest.mark.parametrize("count", [1, 2, 3, 5, 10, 20, 100])
async def test_reader_writer(count, aio_file_maker, temp_file, uuid):
    r_file = await aio_file_maker(temp_file, "rb")
    w_file = await aio_file_maker(temp_file, "wb")

    chunk_size = 16
    payload = os.urandom(chunk_size * count)
    writer = Writer(w_file)

    await writer(payload)
    await w_file.fsync()

    buff = BytesIO(payload)

    async for chunk in Reader(r_file, chunk_size=chunk_size):
        if not chunk:
            break

        assert chunk == buff.read(chunk_size)

    assert not buff.read()


@pytest.mark.parametrize("count", [1, 2, 3, 5, 10, 20, 100, 1000])
async def test_reader_writer2(count, aio_file_maker, temp_file, uuid):
    r_file = await aio_file_maker(temp_file, "r")
    w_file = await aio_file_maker(temp_file, "w")

    writer = Writer(w_file)

    for _ in range(count):
        await writer(uuid)

    await w_file.fsync()

    async for chunk in Reader(r_file, chunk_size=len(uuid)):
        assert chunk == uuid


async def test_parallel_writer(aio_file_maker, temp_file, uuid):
    w_file = await aio_file_maker(temp_file, "w")
    r_file = await aio_file_maker(temp_file, "r")

    futures = list()

    for i in range(1000):
        futures.append(w_file.write(uuid, i * len(uuid)))

    shuffle(futures)

    await asyncio.gather(*futures)
    await w_file.fsync()

    count = 0
    async for chunk in Reader(r_file, chunk_size=len(uuid)):
        assert chunk == uuid
        count += 1

    assert count == 1000


@pytest.mark.parametrize("count", [1, 2, 3, 5, 10, 20, 100, 1000])
async def test_parallel_writer_ordering(
    count, aio_file_maker, temp_file, uuid,
):
    w_file = await aio_file_maker(temp_file, "wb")
    r_file = await aio_file_maker(temp_file, "rb")

    chunk_size = 1024

    data = os.urandom(chunk_size * count)

    futures = list()

    for idx, chunk in enumerate(split_by(data, chunk_size)):
        futures.append(w_file.write(chunk, idx * chunk_size))

    shuffle(futures)

    await asyncio.gather(*futures)
    await w_file.fsync()

    result: bytes = b""

    async for chunk in Reader(r_file, chunk_size=chunk_size):
        result += cast(bytes, chunk)

    assert data == result


async def test_non_existent_file_ctx(aio_file_maker):
    with pytest.raises(FileNotFoundError):
        async with aio_file_maker("/c/windows/NonExistent.file", "r"):
            pass


async def test_sequential_open(aio_file_maker, temp_file):
    file = aio_file_maker(temp_file, "r")
    try:
        assert isinstance(await file.open(), int)
        assert await file.open() is None
    finally:
        await file.close()

    with pytest.raises(asyncio.InvalidStateError):
        await file.open()


async def test_parallel_open(aio_file_maker, temp_file):
    file = aio_file_maker(temp_file, "r")
    try:
        # open() returns a file descriptor if it has opened a file, otherwise
        # None. Only one of multiple parallel calls should actually open a file.
        opens = await asyncio.gather(*(file.open() for _ in range(3)))
        assert len([fd for fd in opens if fd is not None]) == 1
    finally:
        await file.close()


async def test_line_reader(aio_file_maker, temp_file, uuid):
    afp = await aio_file_maker(temp_file, "w+")

    writer = Writer(afp)

    max_length = 1000
    chunk = b64encode(os.urandom(max_length)).decode()
    lines = [chunk[:i] for i in range(max_length)]

    line: str | bytes

    for line in lines:
        await writer(line)
        await writer("\n")

    await afp.fsync()
    read_lines = []

    async for line in LineReader(afp):
        read_lines.append(line[:-1])

    def hash_data(data_lines):
        return hashlib.md5("\n".join(data_lines).encode()).hexdigest()

    assert hash_data(read_lines) == hash_data(lines)


@pytest.mark.parametrize("size", [1, 2, 3, 5, 10, 20, 100, 1000, 2000, 5000])
async def test_line_reader_one_line(size, aio_file_maker, temp_file):
    afp = await aio_file_maker(temp_file, "w+")

    writer = Writer(afp)

    payload = " ".join(uuid4().hex for _ in range(size))

    await writer(payload)

    read_lines = []

    async for line in LineReader(afp):
        read_lines.append(line)

    assert payload == read_lines[0]


async def test_truncate(aio_file_maker, temp_file):
    afp = await aio_file_maker(temp_file, "w+")

    await afp.write("hello")
    await afp.fsync()

    assert (await afp.read()) == "hello"

    await afp.truncate(0)

    assert (await afp.read()) == ""


@pytest.mark.parametrize("size", [1, 2, 3, 5, 10, 20, 100, 1000, 2000, 5000])
async def test_modes(size, aio_file_maker, tmpdir):
    tmpfile = tmpdir.join("test.txt")

    async with aio_file_maker(tmpfile, "w") as afp:
        await afp.write("foo")
        await afp.fsync()

    async with aio_file_maker(tmpfile, "r") as afp:
        assert await afp.read() == "foo"

    async with aio_file_maker(tmpfile, "a+") as afp:
        assert await afp.read() == "foo"

    async with aio_file_maker(tmpfile, "r+") as afp:
        assert await afp.read() == "foo"

    data = dict((str(i), i)for i in range(size))

    tmpfile = tmpdir.join("test.json")
    async with aio_file_maker(tmpfile, "w") as afp:
        await afp.write(json.dumps(data, indent=1))

    async with aio_file_maker(tmpfile, "r") as afp:
        result = json.loads(await afp.read())

    assert result == data


async def test_unicode_reader(aio_file_maker, temp_file):
    async with aio_file_maker(temp_file, "w+") as afp:
        await afp.write("한글")

    async with aio_file_maker(temp_file, "r") as afp:
        reader = Reader(afp, chunk_size=1)
        assert await reader.read_chunk() == "한"
        assert await reader.read_chunk() == "글"


async def test_unicode_writer(aio_file_maker, temp_file):
    async with aio_file_maker(temp_file, "w+") as afp:
        writer = Writer(afp)
        await writer("한")
        await writer("글")

    async with aio_file_maker(temp_file, "r") as afp:
        reader = Reader(afp, chunk_size=1)
        assert await reader.read_chunk() == "한"
        assert await reader.read_chunk() == "글"


@pytest.mark.parametrize(("mode", "data"), [("w+", ""), ("wb+", b"")])
async def test_write_read_nothing(aio_file_maker, temp_file, mode, data):
    async with aio_file_maker(temp_file, mode) as afp:
        assert await afp.write(data) == 0
        assert await afp.read() == data


async def test_partial_writes(temp_file, event_loop):
    ctx = Mock(caio.AbstractContext)
    ctx.loop = event_loop
    ctx.fdsync = CoroutineMock(return_value=None)
    ctx.write = CoroutineMock(side_effect=asyncio.InvalidStateError)

    async with AIOFile(temp_file, "w", context=ctx) as afp:
        # 1. writing first three bytes then last four
        return_iter = iter((3, 4))
        ctx.write.side_effect = lambda *_, **__: next(return_iter)
        await afp.write("aiofile", offset=0)
        # 2. then writing 12 bytes then one and the last 6.
        return_iter = iter((12, 1, 6))
        ctx.write.side_effect = lambda *_, **__: next(return_iter)
        await afp.write("test_partial_writes", offset=8)

        # compare all chunks against expected
        assert ctx.write.call_args_list == [
            # 1
            call(b"aiofile", afp.fileno(), 0),
            call(b"file", afp.fileno(), 3),
            # 2
            call(b"test_partial_writes", afp.fileno(), 8),
            call(b"_writes", afp.fileno(), 20),
            call(b"writes", afp.fileno(), 21),
        ]


async def test_write_returned_negative(temp_file, event_loop):
    ctx = Mock(caio.AbstractContext)
    ctx.loop = event_loop
    ctx.fdsync = CoroutineMock(return_value=None)
    ctx.write = CoroutineMock(side_effect=asyncio.InvalidStateError)

    async with AIOFile(temp_file, "w", context=ctx) as afp:
        return_iter = iter((3, -27))
        ctx.write.side_effect = lambda *_, **__: next(return_iter)
        with pytest.raises(OSError) as raises:
            await afp.write("aiofile")
        assert raises.value.errno == 27
        assert raises.value.filename == temp_file

        ctx.write.reset_mock(side_effect=True)
        ctx.write.return_value = -122
        with pytest.raises(OSError) as raises:
            await afp.write("aiofile")
        assert raises.value.errno == 122
        assert raises.value.filename == temp_file


async def test_write_returned_zero(temp_file, event_loop):
    ctx = Mock(caio.AbstractContext)
    ctx.loop = event_loop
    ctx.fdsync = CoroutineMock(return_value=None)
    ctx.write = CoroutineMock(side_effect=asyncio.InvalidStateError)

    async with AIOFile(temp_file, "w", context=ctx) as afp:
        return_iter = iter((3, 0))
        ctx.write.side_effect = lambda *_, **__: next(return_iter)
        with pytest.raises(RuntimeError, match="Write operation returned 0"):
            await afp.write("aiofile")

        ctx.write.reset_mock(side_effect=True)
        ctx.write.return_value = 0
        with pytest.raises(RuntimeError, match="Write operation returned 0"):
            await afp.write("aiofile")


async def test_text_io_wrapper(aio_file_maker, temp_file):
    async with aio_file_maker(temp_file, "w+") as afp:
        data = "💾💀"
        await afp.write(data * 5)

    with open(temp_file, "a+", encoding="utf-8") as fp:
        assert not fp.read(1)
        fp.seek(0)
        assert fp.read() == data * 5

        fp.seek(0)
        assert fp.read(1) == "💾"
        assert fp.tell() == 4

    async with TextFileWrapper(aio_file_maker(temp_file, "a+")) as fp:
        assert not await fp.read(1)
        fp.seek(0)
        chunk = await fp.read()
        assert chunk == data * 5

        fp.seek(0)
        assert await fp.read(1) == "💾"
        assert fp.tell() == 4


async def test_binary_io_wrapper(aio_file_maker, temp_file):
    async with aio_file_maker(temp_file, "wb+") as afp:
        data = b"\x01\x02\x03"
        await afp.write(data * 32)

    with open(temp_file, "ab+") as fp:
        assert not fp.read(1)
        fp.seek(0)
        assert fp.read() == data * 32

        fp.seek(0)
        assert fp.read(1) == b"\x01"
        assert fp.tell() == 1

    async with BinaryFileWrapper(aio_file_maker(temp_file, "ab+")) as fp:
        assert not await fp.read(1)
        fp.seek(0)
        assert await fp.read() == data * 32

        fp.seek(0)
        assert await fp.read(1) == b"\x01"
        assert fp.tell() == 1


async def test_async_open(async_open, temp_file):
    async with async_open(temp_file, "wb+") as afp:
        data = b"\x01\x02\x03" + "🦠📱".encode()
        await afp.write(data * 32)

    assert isinstance(async_open(temp_file, "ab+"), BinaryFileWrapper)

    async with async_open(temp_file, "ab+") as fp:
        assert not await fp.read(1)
        fp.seek(3)
        assert await fp.read(8) == "🦠📱".encode()

    assert isinstance(async_open(temp_file, "a+"), TextFileWrapper)

    async with async_open(temp_file, "a+") as fp:
        assert not await fp.read(1)
        fp.seek(3)
        assert await fp.read(2) == "🦠📱"


async def test_async_open_unicode(aio_file_maker, async_open, temp_file):
    async with aio_file_maker(temp_file, "w+") as afp:
        data = "🏁💾🏴‍☠️"
        await afp.write(data)

    async with async_open(temp_file, "a+") as afp:
        with open(temp_file, "a+", encoding="utf-8") as fp:
            assert not await afp.read(1)
            assert not fp.read(1)

            afp.seek(0)
            fp.seek(0)

            assert await afp.read(1) == fp.read(1)
            assert await afp.read(1) == fp.read(1)
            assert await afp.read(1) == fp.read(1)
            assert await afp.read(1) == fp.read(1)

            afp.seek(0)
            fp.seek(0)

            assert await afp.read(3) == fp.read(3)
            assert afp.tell() == fp.tell()


async def test_async_open_readline(aio_file_maker, async_open, temp_file):
    async with aio_file_maker(temp_file, "w+") as afp:
        data = "Hello\nworld\n" + ("h" * 10000)
        await afp.write(data)

    async with async_open(temp_file, "a+") as afp:
        with open(temp_file, "a+", encoding="utf-8") as fp:
            afp.seek(0)
            fp.seek(0)

            assert await afp.readline() == fp.readline()
            assert await afp.readline() == fp.readline()
            assert await afp.readline() == fp.readline()

    async with async_open(temp_file, "ab+") as afp:
        with open(temp_file, "ab+") as fp:
            assert not await afp.read(1)
            assert not fp.read(1)

            afp.seek(0)
            fp.seek(0)

            assert await afp.readline() == fp.readline()
            assert await afp.readline() == fp.readline()
            assert await afp.readline() == fp.readline()


async def test_async_open_fp(async_open, tmp_path: Path):
    data = "Hello\nworld\n"

    with open(tmp_path / "file.txt", "w+") as fp:
        async with async_open(fp) as afp:
            await afp.write(data)
            afp.seek(0)
            assert await afp.read() == data
            assert not afp.file.mode.binary

        assert not fp.closed

    with open(tmp_path / "file.txt", "rb") as fp:
        async with async_open(fp) as afp:
            assert afp.file.mode.binary

        assert fp.read().decode() == data


@pytest.mark.parametrize("sizes", [[1, 2], [2, 10], [10, 20], [100, 500]])
async def test_async_open_line_iter(sizes, async_open, tmp_path: Path):
    async with async_open(tmp_path / "file.txt", "w+") as afp:
        for i in range(*sizes):
            await afp.write(str(i))
            await afp.write("\n")

        afp.seek(0)
        idx = sizes[0]
        async for line in afp:
            assert line.endswith("\n")
            assert int(line.strip()) == idx
            idx += 1

    async with async_open(tmp_path / "file.bin", "wb+") as afp:
        for i in range(*sizes):
            await afp.write(str(i).encode())
            await afp.write(b"\n")

        afp.seek(0)
        idx = sizes[0]
        async for line in afp:
            assert line.endswith(b"\n")
            assert int(line.decode().strip()) == idx
            idx += 1


@pytest.mark.parametrize("size", [1, 2, 3, 5, 10, 20, 100, 1000, 2000, 5000])
async def test_async_open_iter_chunked(size, async_open, tmp_path: Path):
    src_path = tmp_path / "src.txt"
    dst_path = tmp_path / "dst.txt"

    async with async_open(src_path, "w") as afp:
        for i in range(0, size):
            await afp.write(str(i))
            await afp.write("\n")

    async with async_open(src_path, "r") as src, \
               async_open(dst_path, "w") as dest:
        async for chunk in src.iter_chunked():
            await dest.write(chunk)

    assert src_path.stat().st_size == dst_path.stat().st_size

    def hash_file(path):
        with open(path, "rb") as fp:
            hasher = hashlib.md5()
            chunk = fp.read(65535)
            while chunk:
                hasher.update(chunk)
                chunk = fp.read(65535)
            return hasher.hexdigest()

    assert hash_file(src_path) == hash_file(dst_path)


async def test_open_non_existent_file_with_append(
    async_open, tmp_path: Path,
):
    tmp_fpath = tmp_path / "numbers.txt"

    async with async_open(tmp_fpath, "a+") as afp:
        for i in range(10):
            await afp.write(str(i))
            await afp.write("\n")

    async with async_open(tmp_fpath, "a+") as afp:
        for i in range(10, 20):
            await afp.write(str(i))
            await afp.write("\n")

    numbers = []
    async with async_open(tmp_fpath, "r") as afp:
        async for line in afp:
            numbers.append(int(line.strip()))

    assert numbers == list(range(20))


async def test_clone(
    async_open, tmp_path: Path,
):
    tmp_fpath = tmp_path / "test.txt"

    async with async_open(tmp_fpath, "w") as afp:
        for i in range(1000, 1003):
            await afp.write(str(i))
            await afp.write("\n")

    async with async_open(tmp_fpath, "r") as afp:
        assert await afp.read(5) == "1000\n"

        async with clone(afp) as cloned:
            assert await cloned.read(5) == "1000\n"

            async with clone(afp) as cloned2:
                assert await cloned2.read(5) == "1000\n"

                assert await cloned.read(5) == await afp.read(5) == "1001\n"


async def test_clone_close(
    async_open, tmp_path: Path,
):
    tmp_fpath = tmp_path / "test.txt"

    async with async_open(tmp_fpath, "w") as afp:
        for i in range(10):
            await afp.write(str(i))
            await afp.write("\n")

    async with async_open(tmp_fpath, "r") as afp:
        await afp.read(10)

        afp_clone = await clone(afp)
        await afp_clone.close()

        assert await afp.read()

    async with async_open(tmp_fpath, "r") as afp:
        await afp.read(10)

        async with clone(afp):
            assert await afp.read()



import asyncio
import collections.abc
import io
import os
from abc import ABC, abstractmethod
from pathlib import Path
from types import MappingProxyType
from typing import Any, Generator, Generic, Optional, Tuple, TypeVar, Union

from .aio import AIOFile, FileIOType


ENCODING_MAP = MappingProxyType({
    "utf-8": 4,
    "utf-16": 8,
    "UTF-8": 4,
    "UTF-16": 8,
})


async def unicode_reader(
    afp: AIOFile, chunk_size: int, offset: int, encoding: str = "utf-8",
) -> Tuple[int, str]:

    if chunk_size < 0:
        chunk_bytes = await afp.read_bytes(-1, offset)
        return len(chunk_bytes), chunk_bytes.decode(encoding=encoding)

    last_error = None
    for retry in range(ENCODING_MAP.get(encoding, 4)):
        chunk_bytes = await afp.read_bytes(chunk_size + retry, offset)
        try:
            chunk = chunk_bytes.decode(encoding=encoding)
            break
        except UnicodeDecodeError as e:
            last_error = e
    else:
        raise last_error    # type: ignore

    chunk_size = len(chunk_bytes)

    return chunk_size, chunk


class Reader(collections.abc.AsyncIterable):
    __slots__ = "_chunk_size", "__offset", "file", "__lock", "encoding"

    CHUNK_SIZE = 32 * 1024

    def __init__(
        self, aio_file: AIOFile, offset: int = 0,
        chunk_size: int = CHUNK_SIZE,
    ):

        self.__lock = asyncio.Lock()
        self.__offset = int(offset)

        self._chunk_size = int(chunk_size)
        self.file = aio_file
        self.encoding = self.file.encoding

    async def read_chunk(self) -> Union[str, bytes]:
        async with self.__lock:
            if self.file.mode.binary:
                chunk = await self.file.read_bytes(
                    self._chunk_size, self.__offset,
                )   # type: Union[str, bytes]
                chunk_size = len(chunk)
            else:
                chunk_size, chunk = await unicode_reader(
                    self.file, self._chunk_size, self.__offset,
                    encoding=self.encoding,
                )
        self.__offset += chunk_size
        return chunk

    async def __anext__(self) -> Union[str, bytes]:
        chunk = await self.read_chunk()

        if not chunk:
            raise StopAsyncIteration(chunk)

        return chunk

    def __aiter__(self) -> "Reader":
        return self


class Writer:
    __slots__ = "__chunk_size", "__offset", "__aio_file", "__lock"

    def __init__(self, aio_file: AIOFile, offset: int = 0):
        self.__offset = int(offset)
        self.__aio_file = aio_file
        self.__lock = asyncio.Lock()

    async def __call__(self, data: Union[str, bytes]) -> None:
        async with self.__lock:
            if isinstance(data, str):
                data = self.__aio_file.encode_bytes(data)

            await self.__aio_file.write_bytes(data, self.__offset)
            self.__offset += len(data)


class LineReader(collections.abc.AsyncIterable):
    CHUNK_SIZE = 4192

    def __init__(
        self, aio_file: AIOFile, offset: int = 0,
        chunk_size: int = CHUNK_SIZE, line_sep: str = "\n",
    ):
        self.__reader = Reader(aio_file, chunk_size=chunk_size, offset=offset)

        self._buffer: Any = (
            io.BytesIO() if aio_file.mode.binary else io.StringIO()
        )

        self.linesep: Any = (
            aio_file.encode_bytes(line_sep)
            if aio_file.mode.binary
            else line_sep
        )

    async def readline(self) -> Union[str, bytes]:
        while True:
            line = self._buffer.readline()
            if line and line.endswith(self.linesep):
                return line

            buffer_remainder = line + self._buffer.read()
            self._buffer.truncate(0)
            self._buffer.seek(0)

            # No line in buffer, read more data
            chunk = await self.__reader.read_chunk()
            if not chunk:
                # No more data, return any remaining content in the buffer
                return buffer_remainder
            # Write remaining + new data back to buffer for next iteration
            self._buffer.write(buffer_remainder)
            self._buffer.write(chunk)
            self._buffer.seek(0)

    async def __anext__(self) -> Union[bytes, str]:
        line = await self.readline()

        if not line:
            # We are finished, close the buffer and raise StopAsyncIteration
            self._buffer.close()
            raise StopAsyncIteration(line)

        return line

    def __aiter__(self) -> "LineReader":
        return self


class FileIOWrapperBase(ABC):
    _READLINE_CHUNK_SIZE = 4192

    def __init__(self, afp: AIOFile, *, offset: int = 0):
        self._offset = offset
        self._lock = asyncio.Lock()
        self.file = afp

        if self.file.mode.appending:
            try:
                self._offset = os.stat(afp.name).st_size
            except FileNotFoundError:
                self._offset = 0

    @abstractmethod
    async def read(self, length: int = -1) -> Any:
        raise NotImplementedError

    @abstractmethod
    async def write(self, data: Any) -> int:
        raise NotImplementedError

    @abstractmethod
    async def readline(
        self, size: int = -1, newline: Any = ...,
    ) -> Union[str, bytes]:
        raise NotImplementedError

    def seek(self, offset: int) -> None:
        self._offset = offset

    def tell(self) -> int:
        return self._offset

    async def flush(self, sync_metadata: bool = False) -> None:
        if sync_metadata:
            await self.file.fsync()
        else:
            await self.file.fdsync()

    async def close(self) -> None:
        await self.file.close()

    def __await__(self) -> Generator[None, None, "FileIOWrapperBase"]:
        yield from self.file.__await__()
        return self

    async def __aenter__(self) -> "FileIOWrapperBase":
        await self.file.open()
        return self

    async def __aexit__(self, *_: Any) -> None:
        await self.close()

    def __aiter__(self) -> LineReader:
        return LineReader(self.file)

    def iter_chunked(self, chunk_size: int = Reader.CHUNK_SIZE) -> Reader:
        return Reader(self.file, chunk_size=chunk_size, offset=self._offset)


class BinaryFileWrapper(FileIOWrapperBase):
    def __init__(self, afp: AIOFile):
        if not afp.mode.binary:
            raise ValueError("Expected file in binary mode")
        super().__init__(afp)

    async def __read(self, length: int) -> bytes:
        data = await self.file.read_bytes(length, self._offset)
        self._offset += len(data)
        return data

    async def read(self, length: int = -1) -> bytes:
        async with self._lock:
            return await self.__read(length)

    async def write(self, data: bytes) -> int:
        async with self._lock:
            operation = self.file.write_bytes(data, self._offset)
            self._offset += len(data)
        await operation
        return len(data)

    async def readline(self, size: int = -1, newline: bytes = b"\n") -> bytes:
        async with self._lock:
            offset = self._offset
            with io.BytesIO() as fp:
                while True:
                    chunk = await self.__read(self._READLINE_CHUNK_SIZE)

                    if chunk:
                        if newline not in chunk:
                            fp.write(chunk)
                            continue

                        fp.write(chunk)

                    if 0 < size <= fp.tell():
                        fp.seek(size)
                        fp.truncate(size)
                        return fp.getvalue()

                    fp.seek(0)
                    line = fp.readline()
                    self._offset = offset + fp.tell()
                    return line


class TextFileWrapper(FileIOWrapperBase):
    def __init__(self, afp: AIOFile):
        if afp.mode.binary:
            raise ValueError("Expected file in text mode")
        super().__init__(afp)
        self.encoding = self.file.encoding

    async def __read(self, length: int) -> str:
        chunk_size = 0
        offset = self._offset
        chunk = ""
        while length < 0 or length > len(chunk):
            part_offset, part = await unicode_reader(
                self.file, length, offset, self.encoding,
            )

            if not part:
                break

            chunk += part
            offset += part_offset

        if chunk_size > length > 0:
            chunk = chunk[:length]
            offset = length

        self._offset = offset
        return chunk

    async def read(self, length: int = -1) -> str:
        async with self._lock:
            return await self.__read(length)

    async def write(self, data: str) -> int:
        async with self._lock:
            data_bytes = data.encode(self.encoding)
            operation = self.file.write_bytes(data_bytes, self._offset)
            self._offset += len(data_bytes)

        await operation
        return len(data_bytes)

    async def readline(self, size: int = -1, newline: str = "\n") -> str:
        async with self._lock:
            offset = self._offset
            with io.StringIO() as fp:
                while True:
                    chunk = await self.__read(self._READLINE_CHUNK_SIZE)

                    if chunk:
                        if newline not in chunk:
                            fp.write(chunk)
                            continue

                        fp.write(chunk)

                    if 0 < size <= fp.tell():
                        fp.seek(size)
                        fp.truncate(size)
                        return fp.getvalue()

                    fp.seek(0)
                    line = fp.readline()
                    self._offset = offset + len(
                        line.encode(encoding=self.encoding),
                    )
                    return line


def async_open(
    file_specifier: Union[str, Path, FileIOType],
    mode: str = "r", *args: Any, **kwargs: Any,
) -> Union[BinaryFileWrapper, TextFileWrapper]:
    if isinstance(file_specifier, (str, Path)):
        afp = AIOFile(str(file_specifier), mode, *args, **kwargs)
    else:
        if args:
            raise ValueError("Arguments denied when IO[Any] opening.")
        afp = AIOFile.from_fp(file_specifier, **kwargs)

    if not afp.mode.binary:
        return TextFileWrapper(afp)

    return BinaryFileWrapper(afp)


T = TypeVar("T", bound=FileIOWrapperBase)


class FileIOCloner(Generic[T]):
    def __init__(self, file: T):
        self.source_afp = file
        self.cloned_afp: Optional[T] = None
        self._lock = asyncio.Lock()

    async def __clone(self) -> T:
        async with self._lock:
            if self.cloned_afp is not None:
                return self.cloned_afp
            self.cloned_afp = self.source_afp.__class__(
                await self.source_afp.file.clone(),
            )
        return self.cloned_afp

    def __await__(self) -> Generator[Any, None, T]:
        return self.__clone().__await__()

    async def __aenter__(self) -> T:
        return await self.__clone()

    async def __aexit__(self, *_: Any) -> None:
        if self.cloned_afp is not None:
            await self.cloned_afp.close()


def clone(afp: FileIOWrapperBase) -> "FileIOCloner[FileIOWrapperBase]":
    return FileIOCloner(afp)


__all__ = (
    "BinaryFileWrapper",
    "FileIOCloner",
    "FileIOWrapperBase",
    "LineReader",
    "Reader",
    "TextFileWrapper",
    "Writer",
    "async_open",
    "clone",
    "unicode_reader",
)



import asyncio
import os
from collections import namedtuple
from concurrent.futures import Executor
from functools import partial
from os import strerror
from pathlib import Path
from typing import (
    Any, Awaitable, BinaryIO, Callable, Dict, Generator, Optional, TextIO,
    TypeVar, Union, cast,
)
from weakref import finalize

import caio
from caio.asyncio_base import AsyncioContextBase


_T = TypeVar("_T")

AIO_FILE_NOT_OPENED = -1
AIO_FILE_CLOSED = -2

FileIOType = Union[TextIO, BinaryIO]

FileMode = namedtuple(
    "FileMode", (
        "readable",
        "writable",
        "plus",
        "appending",
        "created",
        "flags",
        "binary",
    ),
)


def parse_mode(mode: str) -> FileMode:    # noqa: C901
    """ Rewritten from `cpython fileno`_

    .. _cpython fileio: https://bit.ly/2JY2cnp
    """

    flags = os.O_RDONLY

    rwa = False
    writable = False
    readable = False
    plus = False
    appending = False
    created = False
    binary = False

    for m in mode:
        if m == "x":
            rwa = True
            created = True
            writable = True
            flags |= os.O_EXCL | os.O_CREAT

        if m == "r":
            if rwa:
                raise Exception("Bad mode")

            rwa = True
            readable = True

        if m == "w":
            if rwa:
                raise Exception("Bad mode")

            rwa = True
            writable = True

            flags |= os.O_CREAT | os.O_TRUNC

        if m == "a":
            if rwa:
                raise Exception("Bad mode")
            rwa = True
            writable = True
            appending = True
            flags |= os.O_CREAT | os.O_APPEND

        if m == "+":
            if plus:
                raise Exception("Bad mode")
            readable = True
            writable = True
            plus = True

        if m == "b":
            binary = True
            if hasattr(os, "O_BINARY"):
                flags |= os.O_BINARY

    if readable and writable:
        flags |= os.O_RDWR

    elif readable:
        flags |= os.O_RDONLY
    else:
        flags |= os.O_WRONLY

    return FileMode(
        readable=readable,
        writable=writable,
        plus=plus,
        appending=appending,
        created=created,
        flags=flags,
        binary=binary,
    )


class AIOFile:
    _file_obj: Optional[FileIOType]
    _file_obj_owner: bool
    _encoding: str
    _executor: Optional[Executor]
    mode: FileMode
    __open_result: "Optional[asyncio.Future[FileIOType]]"

    def __init__(
        self, filename: Union[str, Path],
        mode: str = "r", encoding: str = "utf-8",
        context: Optional[AsyncioContextBase] = None,
        executor: Optional[Executor] = None,
    ):
        self.__context = context or get_default_context()
        self.__open_result = None

        self._fname = str(filename)
        self._open_mode = mode

        self.mode = parse_mode(mode)

        self._file_obj = None
        self._file_obj_owner = True
        self._encoding = encoding
        self._executor = executor
        self._clone_lock = asyncio.Lock()
        self._clones = 0

    @classmethod
    def from_fp(cls, fp: FileIOType, **kwargs: Any) -> "AIOFile":
        afp = cls(fp.name, fp.mode, **kwargs)
        afp._file_obj = fp
        afp._open_mode = fp.mode
        afp._file_obj_owner = False
        return afp

    def _run_in_thread(
            self, func: "Callable[..., _T]", *args: Any, **kwargs: Any,
    ) -> "asyncio.Future[_T]":
        return self.__context.loop.run_in_executor(
            self._executor, partial(func, *args, **kwargs),
        )

    @property
    def name(self) -> str:
        return self._fname

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self.__context.loop

    @property
    def encoding(self) -> str:
        return self._encoding

    async def open(self) -> Optional[int]:
        if self._file_obj is not None:
            if self._file_obj.closed:
                raise asyncio.InvalidStateError("AIOFile closed")
            return None

        if self.__open_result is None:
            self.__open_result = cast(
                "asyncio.Future[FileIOType]",
                self._run_in_thread(open, self._fname, self._open_mode),
            )
            self._file_obj = await self.__open_result
            self.__open_result = None
            return self._file_obj.fileno()

        await self.__open_result
        return None

    def __repr__(self) -> str:
        return "<AIOFile: %r>" % self._fname

    async def clone(self) -> "AIOFile":
        """Returns self with a ref-count bump; close() is deferred until all
        clones are released."""
        async with self._clone_lock:
            self._clones += 1
            return self

    async def close(self) -> None:
        if self._file_obj is None or not self._file_obj_owner:
            return

        async with self._clone_lock:
            if self._clones > 0:
                self._clones -= 1
                return

        if self.mode.writable:
            await self.fdsync()

        await self._run_in_thread(self._file_obj.close)

    def fileno(self) -> int:
        if self._file_obj is None:
            raise asyncio.InvalidStateError("AIOFile closed")
        return self._file_obj.fileno()

    def __await__(self) -> Generator[None, Any, "AIOFile"]:
        yield from self.open().__await__()
        return self

    async def __aenter__(self) -> "AIOFile":
        await self.open()
        return self

    def __aexit__(self, *args: Any) -> Awaitable[Any]:
        return asyncio.get_event_loop().create_task(self.close())

    async def read(self, size: int = -1, offset: int = 0) -> Union[bytes, str]:
        data = await self.read_bytes(size, offset)
        return data if self.mode.binary else self.decode_bytes(data)

    async def read_bytes(self, size: int = -1, offset: int = 0) -> bytes:
        if size < -1:
            raise ValueError("Unsupported value %d for size" % size)

        if size == -1:
            size = (
                await self._run_in_thread(
                    os.stat,
                    self.fileno(),
                )
            ).st_size

        return await self.__context.read(size, self.fileno(), offset)

    async def write(self, data: Union[str, bytes], offset: int = 0) -> int:
        if self.mode.binary:
            if not isinstance(data, bytes):
                raise ValueError("Data must be bytes in binary mode")
            bytes_data = data
        else:
            if not isinstance(data, str):
                raise ValueError("Data must be str in text mode")
            bytes_data = self.encode_bytes(data)

        return await self.write_bytes(bytes_data, offset)

    def encode_bytes(self, data: str) -> bytes:
        return data.encode(self._encoding)

    def decode_bytes(self, data: bytes) -> str:
        return data.decode(self._encoding)

    async def write_bytes(self, data: bytes, offset: int = 0) -> int:
        data_size = len(data)
        if data_size == 0:
            return 0

        # data can be written partially, see write(2)
        # (https://www.man7.org/linux/man-pages/man2/write.2.html)
        # for example, it can happen when a disk quota or a resource limit
        # is exceeded (in that case subsequent call will return a
        # corresponding error) or write has been interrupted by
        # an incoming signal

        # behaviour here in regard to continue trying to write remaining data
        # corresponds to the behaviour of io.BufferedIOBase
        # (https://docs.python.org/3/library/io.html#io.BufferedIOBase.write)
        # which used by object returned open() with `buffering` argument >= 1
        # (effectively the default)

        written = 0
        while written < data_size:
            res = await self.__context.write(
                data[written:], self.fileno(), offset + written,
            )
            if res == 0:
                raise RuntimeError(
                    "Write operation returned 0", self, offset, written,
                )
            elif res < 0:
                # fix for linux_aio implementation bug in caio<=0.6.1
                # (https://github.com/mosquito/caio/pull/7)
                # and safeguard against future similar issues
                errno = -res
                raise OSError(errno, strerror(errno), self._fname)

            written += res

        return written

    async def fsync(self) -> None:
        return await self.__context.fsync(self.fileno())

    async def fdsync(self) -> None:
        return await self.__context.fdsync(self.fileno())

    def truncate(self, length: int = 0) -> Awaitable[None]:
        return self._run_in_thread(
            os.ftruncate, self.fileno(), length,
        )


ContextStoreType = Dict[asyncio.AbstractEventLoop, caio.AsyncioContext]
DEFAULT_CONTEXT_STORE: ContextStoreType = {}


def create_context(
    max_requests: int = caio.AsyncioContext.MAX_REQUESTS_DEFAULT,
) -> caio.AsyncioContext:
    loop = asyncio.get_event_loop()
    context = caio.AsyncioContext(max_requests, loop=loop)

    def finalizer() -> None:
        context.close()
        DEFAULT_CONTEXT_STORE.pop(loop, None)

    finalize(loop, finalizer)
    DEFAULT_CONTEXT_STORE[loop] = context
    return context


def get_default_context() -> caio.AsyncioContext:
    loop = asyncio.get_event_loop()
    context = DEFAULT_CONTEXT_STORE.get(loop)

    if context is not None:
        return context

    return create_context()
